//
// Font.java - Minimal subset of java.awt.Font
//
// A rather trivial implementation of a Font.
//
// We read a "blittable font" file, which is something from the past I found.
// This is basically just a 1bpp raster with all the code-points drawn into it,
// with a table of their offsets and widths.
// We keep the raster in memory with a byte per pixel, for speed of use later.
// Any undefined/missing characters look like a space.
// We rely upon the space character being defined.
//
// The "blittable fonts" are bundled alongside this code. 
// The facenames we allow are "cr", "scn" and "sf".
// The "Default" font is the "scn" font.
// We have regular forms of these fonts, and bold forms of many.
// We silently ignore requests for italicised fonts.
//
// The descent height isn't in the file, and we have a mapping table.
// The leading isn't in the file, and is simplistically computed.
//
// The FontMetrics class peeks directly at our exposed innards, in order
// to provide information on ascent, descent, string widths, etc..
//
// The Graphics class also peeks within, to draw characters into the bitmap.
//

package nyangau.miniawt;

import java.util.*;
import java.io.*;

public class Font
  {
  public static final int PLAIN  = 0x00;
  public static final int BOLD   = 0x01;
  public static final int ITALIC = 0x02;
  protected int pointsize, style;
  public int rw, rh;  // not normally a part of class Font
  byte[] raster;      // not normally a part of class Font
  public int ascent;  // not normally a part of class Font
  public int descent; // not normally a part of class Font
  public int leading; // not normally a part of class Font
  protected int[] offsets  = new int[0x100];
  public    int[] advances = new int[0x100]; // not normally a part of class Font
  private static int readBytes(InputStream is, byte[] buf, int ntotal)
    throws IOException
    {
    int nsofar = 0;
    while ( nsofar < ntotal )
      {
      int n;
      if ( (n = is.read(buf, nsofar, ntotal-nsofar)) < 0 )
        break;
      nsofar += n;
      }
    return nsofar;
    }
  private static int getWord(byte[] buf, int offset)
    {
    return (buf[offset]&0xff) + ( (buf[offset+1]&0xff) << 8 );
    }
  // File format
  //         2 byte magic (must be 0x8945)
  //         2 byte raster width
  //         2 byte raster height
  //         2 byte bpp (must be 1)
  //   256 * 2 byte offset and 2 byte width
  // rh*(rw/8) bytes of raster
  private void readFontStream(InputStream is)
    throws IOException
    {
    int nheader = 2+2+2+2+256*(2+2);
    byte[] header = new byte[nheader];
    if ( readBytes(is, header, nheader) < nheader )
      throw new IOException("Font header too short");
    if ( getWord(header, 0) != 0x8945 )
      throw new IOException("Font header must have magic number==0x8945");
    rw = getWord(header, 2);
    if ( (rw % 8) != 0 )
      throw new IOException("Font header must have raster width that is a multiple of 8");
    rh = getWord(header, 4);
    if ( getWord(header, 6) != 0x0001 )
      throw new IOException("Font header must have bpp==1");
    descent = 3; // @@@FIXME
    ascent  = rh - descent;
    leading = (ascent+descent)/10;
    int iheader = 2+2+2+2;
    for ( int i = 0; i < 0x100; i++ )
      {
      offsets [i] = getWord(header, iheader  );
      advances[i] = getWord(header, iheader+2);
      iheader += 4;
      }
    for ( int i = 0; i < 0x100; i++ )
      if ( offsets[i] == 0xffff || advances[i] == 0xffff )
        // character missing, draw space instead
        {
        offsets [i] = offsets [32];
        advances[i] = advances[32];
        }
    byte[] r = new byte[rh*rw/8];
    if ( readBytes(is, r, rh*rw/8) < rh*rw/8 )
      throw new IOException("Font raster is too short");
    raster = new byte[rh*rw];
    for ( int i = 0; i < rh*rw; i++ )
      raster[i] = (byte) ( ( r[i>>>3] & (0x80>>>(i&7)) ) != 0 ? 0xff : 0x00 );
    }
  public Font(String face, int style, int pointsize)
    {
    String name = ( face != null ) ? face : "Default";
    if ( name.equals("Default") )
      name = "scn";
    if ( (style & BOLD) != 0 )
      name += "_b";
    else
      name += "_r";
    name += ( "_" + Integer.toString(pointsize) + ".fon" );
    InputStream is = null;
    try
      {
      is = Font.class.getResourceAsStream("/fonts/"+name);
      }
    catch ( NullPointerException e )
      {
      // Can't occur, name is never null
      }
    try
      {
      readFontStream(is);
      }
    catch ( IOException e )
      {
      System.err.println(e.getMessage());
      // Hmm, in trouble now
      }
    this.pointsize = pointsize;
    this.style     = style;
    }
  public int getSize()
    {
    return pointsize;
    }
  public int getStyle()
    {
    return style;
    }
  }
